from pydantic import BaseModel
from typing import Optional


class ConstructEsewaRequest(BaseModel):
    amount: float
    success_url: str
    failure_url: str
    product_code: str
    product_service_charge: float = 0.0
    product_delivery_charge: float = 0.0
    tax_amount: float = 0.0
    total_amount: float
    transaction_uuid: Optional[str] = None


class EsewaRequestPayload(BaseModel):
    amount: float
    success_url: str
    failure_url: str
    product_code: str
    product_service_charge: float = 0.0
    product_delivery_charge: float = 0.0
    tax_amount: float = 0.0
    total_amount: float
    transaction_uuid: str
    signature: str
    signed_field_names: str


class EsewaDecodedResponse(BaseModel):
    transaction_code: str
    status: str
    total_amount: str
    transaction_uuid: str
    product_code: str
    signed_field_names: str
    signature: str



from np_payments.transport.httpx_sync import HTTPTransport
from np_payments.transport.httpx_async import AsyncHTTPTransport
from .scheams.esewa_schemas import ConstructEsewaRequest, EsewaDecodedResponse
from uuid import uuid4
import hmac
import hashlib
import base64
from typing import Optional, Union
from np_payments.base import BasePaymentGateway
import json


class _BaseEsewa(BasePaymentGateway):
    """
    Base class for Esewa payment gateway integration.
    This class provides the foundational methods and properties
    required for implementing the Esewa payment gateway.
    """

    def __init__(
        self,
        base_url: str,
        secret_key: str,
        success_url: str,
        failure_url: str,
        product_code: str,
        transport: Optional[Union[HTTPTransport, AsyncHTTPTransport]] = None,
    ):
        self.base_url = base_url
        self.secret_key = secret_key
        self.success_url = success_url
        self.failure_url = failure_url
        self.product_code = product_code
        self.transport = transport

    def _generate_signature(self, message: str, key: str) -> str:
        key = key.encode("utf-8")
        message = message.encode("utf-8")
        hmac_sha256 = hmac.new(key, message, hashlib.sha256)
        digest = hmac_sha256.digest()
        signature = base64.b64encode(digest).decode("utf-8")
        return signature

    def _build_request_payload(self, data: ConstructEsewaRequest, key: str) -> dict:
        if data.transaction_uuid is None:
            data.transaction_uuid = str(uuid4())

        signature_string = f"total_amount={data.total_amount},transaction_uuid={data.transaction_uuid},product_code={data.product_code}"
        signature = self._generate_signature(message=signature_string, key=key)

        payload = {
            "amount": str(data.amount),
            "success_url": data.success_url,
            "failure_url": data.failure_url,
            "product_code": data.product_code,
            "product_service_charge": data.product_service_charge,
            "product_delivery_charge": data.product_delivery_charge,
            "tax_amount": str(data.tax_amount),
            "total_amount": str(data.total_amount),
            "transaction_uuid": data.transaction_uuid,
            "signature": signature,
            "signed_field_names": "total_amount,transaction_uuid,product_code",
        }

        return payload

    def _build_payload(
        self,
        amount: float,
        success_url: Optional[str],
        failure_url: Optional[str],
        product_code: Optional[str],
        total_amount: Optional[float],
        product_service_charge: float,
        product_delivery_charge: float,
        tax_amount: float,
        transaction_uuid: Optional[str],
    ) -> dict:
        if not total_amount:
            total_amount = (
                amount + product_service_charge + product_delivery_charge + tax_amount
            )

        request = ConstructEsewaRequest(
            amount=amount,
            success_url=success_url or self.success_url,
            failure_url=failure_url or self.failure_url,
            product_code=product_code or self.product_code,
            product_service_charge=product_service_charge,
            product_delivery_charge=product_delivery_charge,
            tax_amount=tax_amount,
            total_amount=total_amount,
            transaction_uuid=transaction_uuid,
        )
        return self._build_request_payload(data=request, key=self.secret_key)

    def initiate_payment(
        self,
        amount: float,
        success_url: str,
        failure_url: str,
        product_code: str,
        total_amount: Optional[float] = None,
        product_service_charge: Optional[float] = 0.0,
        product_delivery_charge: Optional[float] = 0.0,
        tax_amount: Optional[float] = 0.0,
        transaction_uuid: Optional[str] = None,
    ) -> dict:
        if not total_amount:
            total_amount = (
                amount + product_service_charge + product_delivery_charge + tax_amount
            )

        requset = ConstructEsewaRequest(
            amount=amount,
            success_url=success_url,
            failure_url=failure_url,
            product_service_charge=product_service_charge,
            product_delivery_charge=product_delivery_charge,
            tax_amount=tax_amount,
            total_amount=total_amount,
            transaction_uuid=transaction_uuid,
        )
        payload = self._build_request_payload(data=requset, key=self.secret_key)

        response = self.transport.post(data=payload)
        return response

    def validate_payment(self, data: str) -> bool:
        decoded_data = json.loads(base64.b64decode(data).decode("utf-8"))
        data = EsewaDecodedResponse(**decoded_data)
        signature_string = f"transaction_code={data.transaction_code},status={data.status},total_amount={data.total_amount},transaction_uuid={data.transaction_uuid},product_code={data.product_code},signed_field_names={data.signed_field_names}"
        expected_signature = self._generate_signature(
            key=self.secret_key, message=signature_string
        )
        if expected_signature != data.signature:
            return False
        return True



from typing import Optional
from .base_esewa import _BaseEsewa
from np_payments.transport.httpx_async import AsyncHTTPTransport


class AsyncEsewaClient(_BaseEsewa):
    """
    Sync Client class for interacting with the Esewa payment gateway.
    This class extends the base functionality provided by _BaseEsewa
    and can be used to implement specific payment operations.
    """

    def __init__(self, base_url, secret_key, success_url, failure_url, product_code):
        transport = AsyncHTTPTransport(base_url=base_url)
        super().__init__(
            base_url, secret_key, success_url, failure_url, product_code, transport
        )

    async def initiate_payment(
        self,
        amount: float,
        success_url: Optional[str] = None,
        failure_url: Optional[str] = None,
        product_code: Optional[str] = None,
        total_amount: Optional[float] = None,
        product_service_charge: float = 0.0,
        product_delivery_charge: float = 0.0,
        tax_amount: float = 0.0,
        transaction_uuid: Optional[str] = None,
    ) -> dict:
        payload = self._build_payload(
            amount,
            success_url,
            failure_url,
            product_code,
            total_amount,
            product_service_charge,
            product_delivery_charge,
            tax_amount,
            transaction_uuid,
        )
        return await self.transport.post(data=payload)



from abc import ABC, abstractmethod


class BasePaymentGateway(ABC):
    """
    Abstract base class for payment gateways.
    All payment gateways should inherit from this class and implement the required methods.
    """

    @abstractmethod
    def initiate_payment(
        self, amount: float, success_url: str, failuire_url: str
    ) -> bool:
        """
        Process a payment of the specified amount in the specified currency.

        :param amount: The amount to be charged.
        :param currency: The currency in which the payment is made.
        :return: True if the payment was successful, False otherwise.
        """
        pass

    @abstractmethod
    def validate_payment(self, data: str) -> bool:
        """
        Refund a payment based on the transaction ID.

        :param transaction_id: The ID of the transaction to be refunded.
        :return: True if the refund was successful, False otherwise.
        """
        pass



import httpx
from typing import Optional, Union


class AsyncHTTPTransport:
    def __init__(self, base_url: str):
        self.base_url = base_url
        self.client = httpx.AsyncClient(base_url=base_url)

    async def post(self, data: dict, url: Optional[str] = None) -> Union[dict, str]:
        if not url:
            url = self.base_url
        try:
            response = await self.client.post(url, data=data)
            if response.is_redirect:
                location = response.headers.get("Location")
                return location
            return response.json()
        except Exception as e:
            raise RuntimeError(f"HTTP request failed: {e}")

    async def close(self):
        await self.client.aclose()
